expected_index_tables = [
    ("v2_2d_boundary_conditions", "the_geom"),
    ("v2_2d_lateral", "the_geom"),
    ("v2_calculation_point", "the_geom"),
    ("v2_channel", "the_geom"),
    ("v2_connected_pnt", "the_geom"),
    ("v2_connection_nodes", "the_geom"),
    ("v2_connection_nodes", "the_geom_linestring"),
    ("v2_cross_section_location", "the_geom"),
    ("v2_culvert", "the_geom"),
    ("v2_dem_average_area", "the_geom"),
    ("v2_floodfill", "the_geom"),
    ("v2_grid_refinement", "the_geom"),
    ("v2_grid_refinement_area", "the_geom"),
    ("v2_impervious_surface", "the_geom"),
    ("v2_initial_waterlevel", "the_geom"),
    ("v2_levee", "the_geom"),
    ("v2_obstacle", "the_geom"),
    ("v2_pumped_drainage_area", "the_geom"),
    ("v2_surface", "the_geom"),
    ("v2_windshielding", "the_geom"),
]
